@@ -1,4 +1,4 @@
-/* Copyright 2008-2023 Codership Oy <http://www.codership.com>
+/* Copyright 2008-2021 Codership Oy <http://www.codership.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -27,7 +27,6 @@
 #include <cstdlib>
 #include "wsrep_trans_observer.h"
 #include "wsrep_server_state.h"
-#include "wsrep_mysqld.h"
 
 ulong   wsrep_reject_queries;
 
@@ -120,15 +119,8 @@ bool wsrep_on_update (sys_var *self, THD* thd, enum_var_type var_type)
 
       if (wsrep_init())
       {
-        my_error(ER_CANT_OPEN_LIBRARY, MYF(0), tmp, errno, "wsrep_init failed");
-        saved_wsrep_on= false;
-      }
-
-      if (!wsrep_ready_get())
-      {
-        my_error(ER_GALERA_REPLICATION_NOT_SUPPORTED, MYF(0));
-        WSREP_INFO("Failed to start Galera replication. Please check your "
-                   "configuration.");
+        my_error(ER_CANT_OPEN_LIBRARY, MYF(0), tmp, my_error, "wsrep_init failed");
+        //rcode= true;
         saved_wsrep_on= false;
       }
 
@@ -362,12 +354,14 @@ bool wsrep_start_position_check (sys_var *self, THD* thd, set_var* var)
          var->save_result.string_value.length);
   start_pos_buf[var->save_result.string_value.length]= 0;
 
+
   WSREP_DEBUG("SST wsrep_start_position check for new position %s old %s",
-              start_pos_buf, wsrep_start_position);
+	     start_pos_buf, wsrep_start_position);
 
   // Verify the format.
   if (wsrep_start_position_verify(start_pos_buf)) return true;
 
+
   // Give error if position is updated when wsrep is not enabled or
   // provider is not loaded.
   if ((!WSREP_ON || !Wsrep_server_state::instance().is_provider_loaded())
@@ -674,7 +668,7 @@ bool wsrep_cluster_address_update (sys_var *self, THD* thd, enum_var_type type)
   {
     wsrep_create_rollbacker();
     WSREP_DEBUG("Cluster address update creating %ld applier threads running %lu",
-                wsrep_slave_threads, wsrep_running_applier_threads);
+	    wsrep_slave_threads, wsrep_running_applier_threads);
     wsrep_create_appliers(wsrep_slave_threads);
   }
   mysql_mutex_unlock(&LOCK_wsrep_cluster_config);
@@ -778,7 +772,7 @@ static void wsrep_slave_count_change_update ()
 {
   wsrep_slave_count_change = (wsrep_slave_threads - wsrep_running_applier_threads);
   WSREP_DEBUG("Change on slave threads: New %ld old %lu difference %d",
-              wsrep_slave_threads, wsrep_running_applier_threads, wsrep_slave_count_change);
+	  wsrep_slave_threads, wsrep_running_applier_threads, wsrep_slave_count_change);
 }
 
 bool wsrep_slave_threads_update (sys_var *self, THD* thd, enum_var_type type)
@@ -796,30 +790,12 @@ bool wsrep_slave_threads_update (sys_var *self, THD* thd, enum_var_type type)
   if (wsrep_slave_count_change > 0)
   {
     WSREP_DEBUG("Creating %d applier threads, total %ld", wsrep_slave_count_change, wsrep_slave_threads);
-    wsrep_thread_create_failed.store(false, std::memory_order_relaxed);
     res= wsrep_create_appliers(wsrep_slave_count_change, true);
-    mysql_mutex_unlock(&LOCK_global_system_variables);
-    mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
-    // Thread creation and execution is asyncronous, therefore we need
-    // wait them to be started or error produced
-    while (wsrep_running_applier_threads != (ulong)wsrep_slave_threads &&
-           !wsrep_thread_create_failed.load(std::memory_order_relaxed))
-    {
-      my_sleep(1000);
-    }
-
-    mysql_mutex_lock(&LOCK_global_system_variables);
-
-    if (wsrep_thread_create_failed.load(std::memory_order_relaxed)) {
-      wsrep_slave_threads= wsrep_running_applier_threads;
-      return true;
-    }
-
     WSREP_DEBUG("Running %lu applier threads", wsrep_running_applier_threads);
     wsrep_slave_count_change = 0;
   }
-  else
-     mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
+
+  mysql_mutex_unlock(&LOCK_wsrep_slave_threads);
 
   return res;
 }
@@ -861,10 +837,11 @@ bool wsrep_desync_check (sys_var *self, THD* thd, set_var* var)
       return true;
     }
   } else {
+    THD_STAGE_INFO(thd, stage_waiting_flow);
     ret= Wsrep_server_state::instance().provider().resync();
     if (ret != WSREP_OK) {
       WSREP_WARN ("SET resync failed %d for schema: %s, query: %s", ret,
-                  thd->get_db(), thd->query());
+                  thd->get_db(), wsrep_thd_query(thd));
       my_error (ER_CANNOT_USER, MYF(0), "'resync'", thd->query());
       return true;
     }
@@ -981,7 +958,7 @@ bool wsrep_max_ws_size_update(sys_var *self, THD *thd, enum_var_type)
 {
   char max_ws_size_opt[128];
   my_snprintf(max_ws_size_opt, sizeof(max_ws_size_opt),
-              "repl.max_ws_size=%lu", wsrep_max_ws_size);
+              "repl.max_ws_size=%d", wsrep_max_ws_size);
   enum wsrep::provider::status ret= Wsrep_server_state::instance().provider().options(max_ws_size_opt);
   if (ret)
   {
@@ -993,22 +970,6 @@ bool wsrep_max_ws_size_update(sys_var *self, THD *thd, enum_var_type)
 
 bool wsrep_mode_check(sys_var *self, THD* thd, set_var* var)
 {
-  ulonglong new_wsrep_mode= var->save_result.ulonglong_value;
-  ulonglong old_wsrep_mode= wsrep_mode;
-  wsrep_mode= new_wsrep_mode;
-  if (wsrep_check_mode(WSREP_MODE_REPLICATE_MYISAM) ||
-      wsrep_check_mode(WSREP_MODE_REPLICATE_ARIA))
-  {
-    if (!(wsrep_forced_binlog_format == BINLOG_FORMAT_UNSPEC ||
-          wsrep_forced_binlog_format == BINLOG_FORMAT_ROW))
-    {
-      my_message(ER_WRONG_ARGUMENTS, "wsrep_mode=[REPLICATE_MYISAM|REPLICATE_ARIA] "
-                 "can't be enabled if wsrep_forced_binlog != [NONE|ROW]", MYF(0));
-      wsrep_mode= old_wsrep_mode;
-      return true;
-    }
-  }
-  wsrep_mode= old_wsrep_mode;
   return false;
 }
 
@@ -1080,10 +1041,8 @@ static void export_wsrep_status_to_mysql(THD* thd)
 
 #if DYNAMIC
   if (wsrep_status_len != mysql_status_len) {
-    void* tmp= my_realloc(key_memory_WSREP,
-                          mysql_status_vars,
-                          (wsrep_status_len + 1) * sizeof(SHOW_VAR),
-                          MYF(MY_ALLOW_ZERO_PTR));
+    void* tmp= realloc (mysql_status_vars,
+                         (wsrep_status_len + 1) * sizeof(SHOW_VAR));
     if (!tmp) {
 
       sql_print_error ("Out of memory for wsrep status variables."
@@ -1135,15 +1094,6 @@ void wsrep_free_status (THD* thd)
   thd->wsrep_status_vars.clear();
 }
 
-void wsrep_free_status_vars()
-{
-#if DYNAMIC
-  my_free(mysql_status_vars);
-  mysql_status_vars= NULL;
-  mysql_status_len= 0;
-#endif
-}
-
 bool wsrep_gtid_domain_id_update(sys_var* self, THD *thd, enum_var_type)
 {
   WSREP_DEBUG("wsrep_gtid_domain_id_update: %llu",
@@ -1152,65 +1102,3 @@ bool wsrep_gtid_domain_id_update(sys_var* self, THD *thd, enum_var_type)
   return false;
 }
 
-bool wsrep_strict_ddl_update(sys_var *self, THD* thd, enum_var_type var_type)
-{
-  // In case user still sets wsrep_strict_ddl we set new
-  // option to wsrep_mode
-  if (wsrep_strict_ddl)
-    wsrep_mode|= WSREP_MODE_STRICT_REPLICATION;
-  else
-    wsrep_mode&= (~WSREP_MODE_STRICT_REPLICATION);
-  return false;
-}
-
-bool wsrep_replicate_myisam_update(sys_var *self, THD* thd, enum_var_type var_type)
-{
-  // In case user still sets wsrep_replicate_myisam we set new
-  // option to wsrep_mode
-  if (wsrep_replicate_myisam)
-    wsrep_mode|= WSREP_MODE_REPLICATE_MYISAM;
-  else
-    wsrep_mode&= (~WSREP_MODE_REPLICATE_MYISAM);
-  return false;
-}
-
-bool wsrep_replicate_myisam_check(sys_var *self, THD* thd, set_var* var)
-{
-  bool new_replicate_myisam= (bool)var->save_result.ulonglong_value;
-
-  if (new_replicate_myisam &&
-      !(wsrep_forced_binlog_format == BINLOG_FORMAT_UNSPEC ||
-        wsrep_forced_binlog_format == BINLOG_FORMAT_ROW))
-  {
-    my_message(ER_WRONG_ARGUMENTS, "wsrep_mode=REPLICATE_MYISAM can't be enabled "
-               "if wsrep_forced_binlog != [NONE|ROW]", MYF(0));
-    return true;
-  }
-  return false;
-}
-
-bool wsrep_forced_binlog_format_check(sys_var *self, THD* thd, set_var* var)
-{
-  ulonglong new_forced_binlog_format= var->save_result.ulonglong_value;
-
-  if (!(new_forced_binlog_format == BINLOG_FORMAT_UNSPEC ||
-        new_forced_binlog_format == BINLOG_FORMAT_ROW))
-  {
-    if (wsrep_check_mode(WSREP_MODE_BINLOG_ROW_FORMAT_ONLY))
-    {
-      my_message(ER_WRONG_ARGUMENTS, "wsrep_forced_binlog_format=[MIXED|STATEMENT] can't be set "
-                 "if wsrep_mode=BINLOG_ROW_FORMAT_ONLY", MYF(0));
-      return true;
-    }
-
-    if (wsrep_check_mode(WSREP_MODE_REPLICATE_MYISAM) ||
-        wsrep_check_mode(WSREP_MODE_REPLICATE_ARIA))
-    {
-      my_message(ER_WRONG_ARGUMENTS, "wsrep_forced_binlog_format=[MIXED|STATEMENT] can't be set "
-                 "if wsrep_mode=[REPLICATE_MYISAM|REPLICATE_ARIA]", MYF(0));
-      return true;
-    }
-  }
-
-  return false;
-}
