@@ -1,7 +1,7 @@
 /*****************************************************************************
 
 Copyright (c) 1996, 2016, Oracle and/or its affiliates. All Rights Reserved.
-Copyright (c) 2016, 2023, MariaDB Corporation.
+Copyright (c) 2016, 2022, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -33,8 +33,8 @@ Created 4/24/1996 Heikki Tuuri
 #include "dict0boot.h"
 #include "dict0crea.h"
 #include "dict0dict.h"
+#include "dict0mem.h"
 #include "dict0stats.h"
-#include "ibuf0ibuf.h"
 #include "fsp0file.h"
 #include "fts0priv.h"
 #include "mach0data.h"
@@ -215,9 +215,8 @@ dict_startscan_system(
 	mtr_t*		mtr,		/*!< in: the mini-transaction */
 	dict_table_t*	table)		/*!< in: system table */
 {
-  btr_pcur_init(pcur);
-  if (pcur->open_leaf(true, table->indexes.start, BTR_SEARCH_LEAF, mtr) !=
-      DB_SUCCESS)
+  if (btr_pcur_open_at_index_side(true, table->indexes.start, BTR_SEARCH_LEAF,
+                                  pcur, true, 0, mtr) != DB_SUCCESS)
     return nullptr;
   const rec_t *rec;
   do
@@ -613,11 +612,10 @@ dict_sys_tables_type_valid(ulint type, bool not_redundant)
 @param[in]	not_redundant	whether ROW_FORMAT=REDUNDANT is not used
 @return	table flags */
 static
-ulint
-dict_sys_tables_type_to_tf(ulint type, bool not_redundant)
+uint32_t dict_sys_tables_type_to_tf(uint32_t type, bool not_redundant)
 {
 	ut_ad(dict_sys_tables_type_valid(type, not_redundant));
-	ulint	flags = not_redundant ? 1 : 0;
+	uint32_t flags = not_redundant ? 1 : 0;
 
 	/* ZIP_SSIZE, ATOMIC_BLOBS, DATA_DIR, PAGE_COMPRESSION,
 	PAGE_COMPRESSION_LEVEL are the same. */
@@ -655,15 +653,14 @@ dict_sys_tables_rec_read(
 	bool			uncommitted,
 	mtr_t*			mtr,
 	table_id_t*		table_id,
-	ulint*			space_id,
-	ulint*			n_cols,
-	ulint*			flags,
-	ulint*			flags2,
+	uint32_t*		space_id,
+	uint32_t*		n_cols,
+	uint32_t*		flags,
+	uint32_t*		flags2,
 	trx_id_t*		trx_id)
 {
 	const byte*	field;
 	ulint		len;
-	ulint		type;
 	mem_heap_t*	heap = nullptr;
 
 	field = rec_get_nth_field_old(
@@ -721,7 +718,7 @@ dict_sys_tables_rec_read(
 	field = rec_get_nth_field_old(
 		rec, DICT_FLD__SYS_TABLES__TYPE, &len);
 	ut_a(len == 4);
-	type = mach_read_from_4(field);
+	uint32_t type = mach_read_from_4(field);
 
 	/* Handle MDEV-12873 InnoDB SYS_TABLES.TYPE incompatibility
 	for PAGE_COMPRESSED=YES in MariaDB 10.2.2 to 10.2.6.
@@ -812,8 +809,8 @@ dict_sys_tables_rec_read(
 	if (!dict_sys_tables_type_valid(type, not_redundant)) {
 		sql_print_error("InnoDB: Table %.*s in InnoDB"
 				" data dictionary contains invalid flags."
-				" SYS_TABLES.TYPE=" ULINTPF
-				" SYS_TABLES.N_COLS=" ULINTPF,
+				" SYS_TABLES.TYPE=" UINT32PF
+				" SYS_TABLES.N_COLS=" UINT32PF,
 				int(rec_get_field_start_offs(rec, 1)), rec,
 				type, *n_cols);
 err_exit:
@@ -843,8 +840,8 @@ err_exit:
 			sql_print_error("InnoDB: Table %.*s in InnoDB"
 					" data dictionary"
 					" contains invalid flags."
-					" SYS_TABLES.TYPE=" ULINTPF
-					" SYS_TABLES.MIX_LEN=" ULINTPF,
+					" SYS_TABLES.TYPE=" UINT32PF
+					" SYS_TABLES.MIX_LEN=" UINT32PF,
 					int(rec_get_field_start_offs(rec, 1)),
 					rec,
 					type, *flags2);
@@ -867,32 +864,20 @@ err_exit:
 	return READ_OK;
 }
 
-/** @return SELECT MAX(space) FROM sys_tables */
-static uint32_t dict_find_max_space_id(btr_pcur_t *pcur, mtr_t *mtr)
+/** Check each tablespace found in the data dictionary.
+Then look at each table defined in SYS_TABLES that has a space_id > 0
+to find all the file-per-table tablespaces.
+
+In a crash recovery we already have some tablespace objects created from
+processing the REDO log. We will compare the
+space_id information in the data dictionary to what we find in the
+tablespace file. In addition, more validation will be done if recovery
+was needed and force_recovery is not set.
+
+We also scan the biggest space id, and store it to fil_system. */
+void dict_check_tablespaces_and_store_max_id()
 {
-  uint32_t max_space_id= 0;
-
-  for (const rec_t *rec= dict_startscan_system(pcur, mtr, dict_sys.sys_tables);
-       rec; rec= dict_getnext_system_low(pcur, mtr))
-    if (!dict_sys_tables_rec_check(rec))
-    {
-      ulint len;
-      const byte *field=
-        rec_get_nth_field_old(rec, DICT_FLD__SYS_TABLES__SPACE, &len);
-      ut_ad(len == 4);
-      max_space_id= std::max(max_space_id, mach_read_from_4(field));
-    }
-
-  return max_space_id;
-}
-
-/** Check MAX(SPACE) FROM SYS_TABLES and store it in fil_system.
-Open each data file if an encryption plugin has been loaded.
-
-@param spaces  set of tablespace files to open */
-void dict_check_tablespaces_and_store_max_id(const std::set<uint32_t> *spaces)
-{
-	ulint		max_space_id = 0;
+	uint32_t	max_space_id = 0;
 	btr_pcur_t	pcur;
 	mtr_t		mtr;
 
@@ -902,21 +887,15 @@ void dict_check_tablespaces_and_store_max_id(const std::set<uint32_t> *spaces)
 
 	dict_sys.lock(SRW_LOCK_CALL);
 
-	if (!spaces && ibuf.empty
-	    && !encryption_key_id_exists(FIL_DEFAULT_ENCRYPTION_KEY)) {
-		max_space_id = dict_find_max_space_id(&pcur, &mtr);
-		goto done;
-	}
-
 	for (const rec_t *rec = dict_startscan_system(&pcur, &mtr,
 						      dict_sys.sys_tables);
 	     rec; rec = dict_getnext_system_low(&pcur, &mtr)) {
 		ulint		len;
 		table_id_t	table_id;
-		ulint		space_id;
-		ulint		n_cols;
-		ulint		flags;
-		ulint		flags2;
+		uint32_t	space_id;
+		uint32_t	n_cols;
+		uint32_t	flags;
+		uint32_t	flags2;
 
 		/* If a table record is not useable, ignore it and continue
 		on to the next record. Error messages were logged. */
@@ -939,6 +918,14 @@ void dict_check_tablespaces_and_store_max_id(const std::set<uint32_t> *spaces)
 			continue;
 		}
 
+		if (flags2 & DICT_TF2_DISCARDED) {
+			sql_print_information("InnoDB: Ignoring tablespace"
+					      " for %.*s because "
+					      "the DISCARD flag is set",
+					      static_cast<int>(len), field);
+			continue;
+		}
+
 		/* For tables or partitions using .ibd files, the flag
 		DICT_TF2_USE_FILE_PER_TABLE was not set in MIX_LEN
 		before MySQL 5.6.5. The flag should not have been
@@ -951,19 +938,6 @@ void dict_check_tablespaces_and_store_max_id(const std::set<uint32_t> *spaces)
 			continue;
 		}
 
-		if (spaces && spaces->find(uint32_t(space_id))
-                    == spaces->end()) {
-			continue;
-		}
-
-		if (flags2 & DICT_TF2_DISCARDED) {
-			sql_print_information("InnoDB: Ignoring tablespace"
-					      " for %.*s because "
-					      "the DISCARD flag is set",
-					      static_cast<int>(len), field);
-			continue;
-		}
-
 		const span<const char> name{field, len};
 
 		char*	filepath = fil_make_filepath(nullptr, name,
@@ -972,10 +946,8 @@ void dict_check_tablespaces_and_store_max_id(const std::set<uint32_t> *spaces)
 		const bool not_dropped{!rec_get_deleted_flag(rec, 0)};
 
 		/* Check that the .ibd file exists. */
-		if (fil_ibd_open(space_id, dict_tf_to_fsp_flags(flags),
-				 not_dropped
-				 ? fil_space_t::VALIDATE_NOTHING
-				 : fil_space_t::MAYBE_MISSING,
+		if (fil_ibd_open(not_dropped, FIL_TYPE_TABLESPACE,
+				 space_id, dict_tf_to_fsp_flags(flags),
 				 name, filepath)) {
 		} else if (!not_dropped) {
 		} else if (srv_operation == SRV_OPERATION_NORMAL
@@ -998,7 +970,6 @@ void dict_check_tablespaces_and_store_max_id(const std::set<uint32_t> *spaces)
 		ut_free(filepath);
 	}
 
-done:
 	mtr.commit();
 
 	fil_set_max_space_id_if_bigger(max_space_id);
@@ -1345,9 +1316,8 @@ static dberr_t dict_load_columns(dict_table_t *table, unsigned use_uncommitted,
 	mach_write_to_8(table_id, table->id);
 	dfield_set_data(&dfield, table_id, 8);
 	dict_index_copy_types(&tuple, sys_index, 1);
-	pcur.btr_cur.page_cur.index = sys_index;
 
-	dberr_t err = btr_pcur_open_on_user_rec(&tuple,
+	dberr_t err = btr_pcur_open_on_user_rec(sys_index, &tuple, PAGE_CUR_GE,
 						BTR_SEARCH_LEAF, &pcur, &mtr);
 	if (err != DB_SUCCESS) {
 		goto func_exit;
@@ -1476,9 +1446,9 @@ dict_load_virtual_col(dict_table_t *table, bool uncommitted, ulint nth_v_col)
 	dfield_set_data(&dfield[1], vcol_pos, 4);
 
 	dict_index_copy_types(&tuple, sys_virtual_index, 2);
-	pcur.btr_cur.page_cur.index = sys_virtual_index;
 
-	dberr_t err = btr_pcur_open_on_user_rec(&tuple,
+	dberr_t err = btr_pcur_open_on_user_rec(sys_virtual_index, &tuple,
+						PAGE_CUR_GE,
 						BTR_SEARCH_LEAF, &pcur, &mtr);
 	if (err != DB_SUCCESS) {
 		goto func_exit;
@@ -1712,9 +1682,9 @@ static dberr_t dict_load_fields(dict_index_t *index, bool uncommitted,
 	mach_write_to_8(index_id, index->id);
 	dfield_set_data(&dfield, index_id, 8);
 	dict_index_copy_types(&tuple, sys_index, 1);
-	pcur.btr_cur.page_cur.index = sys_index;
 
-	dberr_t error = btr_pcur_open_on_user_rec(&tuple, BTR_SEARCH_LEAF,
+	dberr_t error = btr_pcur_open_on_user_rec(sys_index, &tuple,
+						  PAGE_CUR_GE, BTR_SEARCH_LEAF,
 						  &pcur, &mtr);
 	if (error != DB_SUCCESS) {
 		goto func_exit;
@@ -1970,9 +1940,9 @@ dberr_t dict_load_indexes(dict_table_t *table, bool uncommitted,
 	mach_write_to_8(table_id, table->id);
 	dfield_set_data(&dfield, table_id, 8);
 	dict_index_copy_types(&tuple, sys_index, 1);
-	pcur.btr_cur.page_cur.index = sys_index;
 
-	dberr_t error = btr_pcur_open_on_user_rec(&tuple, BTR_SEARCH_LEAF,
+	dberr_t error = btr_pcur_open_on_user_rec(sys_index, &tuple,
+						  PAGE_CUR_GE, BTR_SEARCH_LEAF,
 						  &pcur, &mtr);
 	if (error != DB_SUCCESS) {
 		goto func_exit;
@@ -2189,13 +2159,9 @@ const char *dict_load_table_low(mtr_t *mtr, bool uncommitted,
                                 const rec_t *rec, dict_table_t **table)
 {
 	table_id_t	table_id;
-	ulint		space_id;
-	ulint		n_cols;
-	ulint		t_num;
-	ulint		flags;
-	ulint		flags2;
+	uint32_t	space_id, t_num, flags, flags2;
+	ulint		n_cols, n_v_col;
 	trx_id_t	trx_id;
-	ulint		n_v_col;
 
 	if (const char* error_text = dict_sys_tables_rec_check(rec)) {
 		*table = NULL;
@@ -2274,10 +2240,22 @@ dict_load_tablespace(
 	/* The tablespace may already be open. */
 	table->space = fil_space_for_table_exists_in_mem(table->space_id,
 							 table->flags);
-	if (table->space || table->file_unreadable) {
+	if (table->space) {
 		return;
 	}
 
+	if (ignore_err >= DICT_ERR_IGNORE_TABLESPACE) {
+		table->file_unreadable = true;
+		return;
+	}
+
+	if (!(ignore_err & DICT_ERR_IGNORE_RECOVER_LOCK)) {
+		ib::error() << "Failed to find tablespace for table "
+			<< table->name << " in the cache. Attempting"
+			" to load the tablespace with space id "
+			<< table->space_id;
+	}
+
 	/* Use the remote filepath if needed. This parameter is optional
 	in the call to fil_ibd_open(). If not supplied, it will be built
 	from the table->name. */
@@ -2293,19 +2271,13 @@ dict_load_tablespace(
 	}
 
 	table->space = fil_ibd_open(
-		table->space_id, dict_tf_to_fsp_flags(table->flags),
-		fil_space_t::VALIDATE_SPACE_ID,
+		2, FIL_TYPE_TABLESPACE, table->space_id,
+		dict_tf_to_fsp_flags(table->flags),
 		{table->name.m_name, strlen(table->name.m_name)}, filepath);
 
 	if (!table->space) {
 		/* We failed to find a sensible tablespace file */
 		table->file_unreadable = true;
-
-		if (!(ignore_err & DICT_ERR_IGNORE_RECOVER_LOCK)) {
-			sql_print_error("InnoDB: Failed to load tablespace "
-					ULINTPF " for table %s",
-					table->space_id, table->name.m_name);
-		}
 	}
 
 	ut_free(filepath);
@@ -2362,12 +2334,11 @@ static dict_table_t *dict_load_table_one(const span<const char> &name,
 	};
 	dfield_set_data(&dfield, name.data(), name.size());
 	dict_index_copy_types(&tuple, sys_index, 1);
-	pcur.btr_cur.page_cur.index = sys_index;
 
 	bool uncommitted = false;
 reload:
 	mtr.start();
-	dberr_t err = btr_pcur_open_on_user_rec(&tuple,
+	dberr_t err = btr_pcur_open_on_user_rec(sys_index, &tuple, PAGE_CUR_GE,
 						BTR_SEARCH_LEAF, &pcur, &mtr);
 
 	if (err != DB_SUCCESS || !btr_pcur_is_on_user_rec(&pcur)) {
@@ -2476,21 +2447,24 @@ corrupted:
 			only to delete the .ibd files. */
 			goto corrupted;
 		} else {
+			const page_id_t page_id{table->space->id, pk->page};
 			mtr.start();
-			bool ok = false;
-			if (buf_block_t* b = buf_page_get(
-				    page_id_t(table->space->id, pk->page),
-				    table->space->zip_size(),
-				    RW_S_LATCH, &mtr)) {
-				switch (mach_read_from_2(FIL_PAGE_TYPE
-							 + b->page.frame)) {
-				case FIL_PAGE_INDEX:
-				case FIL_PAGE_TYPE_INSTANT:
-					ok = true;
-				}
-			}
+			buf_block_t* block = buf_page_get(
+				page_id, table->space->zip_size(),
+				RW_S_LATCH, &mtr);
+			const bool corrupted = !block
+				|| page_get_space_id(block->page.frame)
+				!= page_id.space()
+				|| page_get_page_no(block->page.frame)
+				!= page_id.page_no()
+				|| (mach_read_from_2(FIL_PAGE_TYPE
+						    + block->page.frame)
+				    != FIL_PAGE_INDEX
+				    && mach_read_from_2(FIL_PAGE_TYPE
+							+ block->page.frame)
+				    != FIL_PAGE_TYPE_INSTANT);
 			mtr.commit();
-			if (!ok) {
+			if (corrupted) {
 				goto corrupted;
 			}
 
@@ -2506,17 +2480,19 @@ corrupted:
 		}
 	}
 
+	/* Initialize table foreign_child value. Its value could be
+	changed when dict_load_foreigns() is called below */
+	table->fk_max_recusive_level = 0;
+
 	/* We will load the foreign key information only if
 	all indexes were loaded. */
 	if (!table->is_readable()) {
 		/* Don't attempt to load the indexes from disk. */
 	} else if (err == DB_SUCCESS) {
-		auto i = fk_tables.size();
-		err = dict_load_foreigns(table->name.m_name, nullptr,
+		err = dict_load_foreigns(table->name.m_name, nullptr, false,
 					 0, true, ignore_err, fk_tables);
 
 		if (err != DB_SUCCESS) {
-			fk_tables.erase(fk_tables.begin() + i, fk_tables.end());
 			ib::warn() << "Load table " << table->name
 				<< " failed, the table has missing"
 				" foreign key indexes. Turn off"
@@ -2524,6 +2500,7 @@ corrupted:
 			goto evict;
 		} else {
 			dict_mem_table_fill_foreign_vcol_set(table);
+			table->fk_max_recusive_level = 0;
 		}
 	}
 
@@ -2541,8 +2518,7 @@ corrupted:
 			/* the table->fts could be created in dict_load_column
 			when a user defined FTS_DOC_ID is present, but no
 			FTS */
-			table->fts->~fts_t();
-			table->fts = nullptr;
+			fts_free(table);
 		} else if (fts_optimize_wq) {
 			fts_optimize_add_table(table);
 		} else if (table->can_be_evicted) {
@@ -2558,7 +2534,7 @@ corrupted:
 }
 
 dict_table_t *dict_sys_t::load_table(const span<const char> &name,
-                                     dict_err_ignore_t ignore) noexcept
+                                     dict_err_ignore_t ignore)
 {
   if (dict_table_t *table= find_table(name))
     return table;
@@ -2616,11 +2592,11 @@ dict_load_table_on_id(
 	mach_write_to_8(id_buf, table_id);
 	dfield_set_data(&dfield, id_buf, 8);
 	dict_index_copy_types(&tuple, sys_table_ids, 1);
-	pcur.btr_cur.page_cur.index = sys_table_ids;
 
 	dict_table_t* table = nullptr;
 
-	if (btr_pcur_open_on_user_rec(&tuple, BTR_SEARCH_LEAF, &pcur, &mtr)
+	if (btr_pcur_open_on_user_rec(sys_table_ids, &tuple, PAGE_CUR_GE,
+				      BTR_SEARCH_LEAF, &pcur, &mtr)
 	    == DB_SUCCESS
 	    && btr_pcur_is_on_user_rec(&pcur)) {
 		/*---------------------------------------------------*/
@@ -2723,10 +2699,9 @@ static dberr_t dict_load_foreign_cols(dict_foreign_t *foreign, trx_id_t trx_id)
 
 	dfield_set_data(&dfield, foreign->id, id_len);
 	dict_index_copy_types(&tuple, sys_index, 1);
-	pcur.btr_cur.page_cur.index = sys_index;
 
 	mem_heap_t* heap = nullptr;
-	dberr_t err = btr_pcur_open_on_user_rec(&tuple,
+	dberr_t err = btr_pcur_open_on_user_rec(sys_index, &tuple, PAGE_CUR_GE,
 						BTR_SEARCH_LEAF, &pcur, &mtr);
 	if (err != DB_SUCCESS) {
 		goto func_exit;
@@ -2898,12 +2873,11 @@ dict_load_foreign(
 	};
 	dfield_set_data(&dfield, id.data(), id.size());
 	dict_index_copy_types(&tuple, sys_index, 1);
-	pcur.btr_cur.page_cur.index = sys_index;
 
 	mtr.start();
 
 	mem_heap_t* heap = nullptr;
-	dberr_t err = btr_pcur_open_on_user_rec(&tuple,
+	dberr_t err = btr_pcur_open_on_user_rec(sys_index, &tuple, PAGE_CUR_GE,
 						BTR_SEARCH_LEAF, &pcur, &mtr);
 	if (err != DB_SUCCESS) {
 		goto err_exit;
@@ -3065,6 +3039,8 @@ dict_load_foreigns(
 	const char*		table_name,	/*!< in: table name */
 	const char**		col_names,	/*!< in: column names, or NULL
 						to use table->col_names */
+	bool			uncommitted,	/*!< in: use READ UNCOMMITTED
+						transaction isolation level */
 	trx_id_t		trx_id,		/*!< in: DDL transaction id,
 						or 0 to check
 						recursive load of tables
@@ -3112,9 +3088,8 @@ start_load:
 	mtr.start();
 	dfield_set_data(&dfield, table_name, strlen(table_name));
 	dict_index_copy_types(&tuple, sec_index, 1);
-	pcur.btr_cur.page_cur.index = sec_index;
 
-	dberr_t err = btr_pcur_open_on_user_rec(&tuple,
+	dberr_t err = btr_pcur_open_on_user_rec(sec_index, &tuple, PAGE_CUR_GE,
 						BTR_SEARCH_LEAF, &pcur, &mtr);
 	if (err != DB_SUCCESS) {
 		DBUG_RETURN(err);
@@ -3180,7 +3155,7 @@ loop:
 	/* Load the foreign constraint definition to the dictionary cache */
 
 	err = len < sizeof fk_id
-		? dict_load_foreign(table_name, false, col_names, trx_id,
+		? dict_load_foreign(table_name, uncommitted, col_names, trx_id,
 				    check_recursive, check_charsets,
 				    {fk_id, len}, ignore_err, fk_tables)
 		: DB_CORRUPTION;
@@ -3215,13 +3190,15 @@ next_rec:
 
 load_next_index:
 	mtr.commit();
+	ut_free(pcur.old_rec_buf);
 
 	if ((sec_index = dict_table_get_next_index(sec_index))) {
-		/* Switch to scan index on REF_NAME */
+		/* Switch to scan index on REF_NAME, fk_max_recusive_level
+		already been updated when scanning FOR_NAME index, no need to
+		update again */
 		check_recursive = false;
 		goto start_load;
 	}
 
-	ut_free(pcur.old_rec_buf);
 	DBUG_RETURN(DB_SUCCESS);
 }
