@@ -1,9 +1,16 @@
 /*****************************************************************************
 
 Copyright (c) 1996, 2017, Oracle and/or its affiliates. All rights reserved.
+Copyright (c) 2008, Google Inc.
 Copyright (c) 2009, Percona Inc.
 Copyright (c) 2013, 2022, MariaDB Corporation.
 
+Portions of this file contain modifications contributed and copyrighted by
+Google, Inc. Those modifications are gratefully acknowledged and are described
+briefly in the InnoDB documentation. The contributions by Google are
+incorporated with their permission, and subject to the conditions contained in
+the file COPYING.Google.
+
 Portions of this file contain modifications contributed and copyrighted
 by Percona Inc.. Those modifications are
 gratefully acknowledged and are described briefly in the InnoDB
@@ -90,7 +97,10 @@ Created 2/16/1996 Heikki Tuuri
 #include "btr0pcur.h"
 #include "zlib.h"
 #include "ut0crc32.h"
-#include "log.h"
+
+/** We are prepared for a situation that we have this many threads waiting for
+a transactional lock inside InnoDB. srv_start() sets the value. */
+ulint srv_max_n_threads;
 
 /** Log sequence number at shutdown */
 lsn_t	srv_shutdown_lsn;
@@ -98,8 +108,11 @@ lsn_t	srv_shutdown_lsn;
 /** TRUE if a raw partition is in use */
 ibool	srv_start_raw_disk_in_use;
 
+/** Number of IO threads to use */
+uint	srv_n_file_io_threads;
+
 /** UNDO tablespaces starts with space id. */
-ulint	srv_undo_space_id_start;
+uint32_t srv_undo_space_id_start;
 
 /** TRUE if the server is being started, before rolling back any
 incomplete transactions */
@@ -239,17 +252,7 @@ static dberr_t create_log_file(bool create_new_db, lsn_t lsn,
 	}
 
 	DBUG_PRINT("ib_log", ("After innodb_log_abort_6"));
-	ut_ad(!os_aio_pending_reads());
-	ut_d(mysql_mutex_lock(&buf_pool.flush_list_mutex));
-	ut_ad(!buf_pool.get_oldest_modification(0));
-	ut_d(mysql_mutex_unlock(&buf_pool.flush_list_mutex));
-	/* os_aio_pending_writes() may hold here if some
-	write_io_callback() did not release the slot yet.  However,
-	the page write itself must have completed, because the
-	buf_pool.flush_list is empty. In debug builds, we wait for
-	this to happen, hoping to get a hung process if this
-	assumption does not hold. */
-	ut_d(os_aio_wait_until_no_pending_writes(false));
+	DBUG_ASSERT(!buf_pool.any_io_pending());
 
 	DBUG_EXECUTE_IF("innodb_log_abort_7", return DB_ERROR;);
 	DBUG_PRINT("ib_log", ("After innodb_log_abort_7"));
@@ -260,7 +263,7 @@ static dberr_t create_log_file(bool create_new_db, lsn_t lsn,
 	bool ret;
 	pfs_os_file_t file = os_file_create(
 		innodb_log_file_key, logfile0.c_str(),
-		OS_FILE_CREATE|OS_FILE_ON_ERROR_NO_EXIT,
+		OS_FILE_CREATE|OS_FILE_ON_ERROR_NO_EXIT, OS_FILE_NORMAL,
 		OS_LOG_FILE, srv_read_only_mode, &ret);
 
 	if (!ret) {
@@ -292,6 +295,9 @@ static dberr_t create_log_file(bool create_new_db, lsn_t lsn,
 	log_sys.log.create();
 
 	log_sys.log.open_file(logfile0);
+	if (!fil_system.sys_space->open(create_new_db)) {
+		return DB_ERROR;
+	}
 
 	/* Create a log checkpoint. */
 	mysql_mutex_lock(&log_sys.mutex);
@@ -324,11 +330,8 @@ static dberr_t create_log_file(bool create_new_db, lsn_t lsn,
 		srv_startup_is_before_trx_rollback_phase = false;
 	}
 
-	/* After disabling recv_no_log_write, enable checkpoints
-	in buf_flush_page_cleaner(). This could help to avoid
-	crash during log file resizing */
+	/* Enable checkpoints in buf_flush_page_cleaner(). */
 	recv_sys.recovery_on = false;
-
 	mysql_mutex_unlock(&log_sys.mutex);
 
 	log_make_checkpoint();
@@ -386,16 +389,16 @@ static dberr_t srv_undo_tablespace_create(const char* name)
 		innodb_data_file_key,
 		name,
 		srv_read_only_mode ? OS_FILE_OPEN : OS_FILE_CREATE,
-		OS_DATA_FILE, srv_read_only_mode, &ret);
+		OS_FILE_NORMAL, OS_DATA_FILE, srv_read_only_mode, &ret);
 
 	if (!ret) {
 		if (os_file_get_last_error(false) != OS_FILE_ALREADY_EXISTS
-#ifdef _AIX
+#ifdef UNIV_AIX
 			/* AIX 5.1 after security patch ML7 may have
 			errno set to 0 here, which causes our function
 			to return 100; work around that AIX problem */
 		    && os_file_get_last_error(false) != 100
-#endif
+#endif /* UNIV_AIX */
 		) {
 			ib::error() << "Can't create UNDO tablespace "
 				<< name;
@@ -459,7 +462,7 @@ static dberr_t srv_validate_undo_tablespaces()
 }
 
 /** @return the number of active undo tablespaces (except system tablespace) */
-static ulint trx_rseg_get_n_undo_tablespaces()
+static uint32_t trx_rseg_get_n_undo_tablespaces()
 {
   std::set<uint32_t> space_ids;
   mtr_t mtr;
@@ -471,7 +474,7 @@ static ulint trx_rseg_get_n_undo_tablespaces()
         if (uint32_t space= trx_sysf_rseg_get_space(sys_header, rseg_id))
           space_ids.insert(space);
   mtr.commit();
-  return space_ids.size();
+  return static_cast<uint32_t>(space_ids.size());
 }
 
 /** Open an undo tablespace.
@@ -479,13 +482,13 @@ static ulint trx_rseg_get_n_undo_tablespaces()
 @param[in]	name	tablespace file name
 @param[in]	i	undo tablespace count
 @return undo tablespace identifier
-@retval 0               if file doesn't exist
-@retval ULINT_UNDEFINED if page0 is corrupted */
-static ulint srv_undo_tablespace_open(bool create, const char* name, ulint i)
+@retval 0 on failure */
+static uint32_t srv_undo_tablespace_open(bool create, const char *name,
+                                         uint32_t i)
 {
   bool success;
-  ulint space_id= 0;
-  ulint fsp_flags= 0;
+  uint32_t space_id= 0;
+  uint32_t fsp_flags= 0;
 
   if (create)
   {
@@ -503,70 +506,48 @@ static ulint srv_undo_tablespace_open(bool create, const char* name, ulint i)
   pfs_os_file_t fh= os_file_create(innodb_data_file_key, name, OS_FILE_OPEN |
                                    OS_FILE_ON_ERROR_NO_EXIT |
                                    OS_FILE_ON_ERROR_SILENT,
-                                   OS_DATA_FILE,
+                                   OS_FILE_AIO, OS_DATA_FILE,
                                    srv_read_only_mode, &success);
 
   if (!success)
     return 0;
 
-  ulint n_retries = 5;
   os_offset_t size= os_file_get_size(fh);
   ut_a(size != os_offset_t(-1));
-  page_t *apage= nullptr;
-  const page_t *page= nullptr;
 
   if (!create)
   {
-    apage= static_cast<byte*>(aligned_malloc(srv_page_size, srv_page_size));
-undo_retry:
-    if (os_file_read(IORequestRead, fh, apage, 0, srv_page_size, nullptr) !=
-        DB_SUCCESS)
+    page_t *page= static_cast<byte*>(aligned_malloc(srv_page_size,
+                                                    srv_page_size));
+    dberr_t err= os_file_read(IORequestRead, fh, page, 0, srv_page_size);
+    if (err != DB_SUCCESS)
     {
 err_exit:
-      if (n_retries && srv_operation == SRV_OPERATION_BACKUP)
-      {
-        sql_print_information("InnoDB: Retrying to read undo "
-                              "tablespace %s", name);
-        n_retries--;
-        goto undo_retry;
-      }
-      sql_print_error("InnoDB: Unable to read first page of file %s", name);
-      aligned_free(apage);
-      return ULINT_UNDEFINED;
+      ib::error() << "Unable to read first page of file " << name;
+      aligned_free(page);
+      return err;
     }
 
-    page= apage;
-    DBUG_EXECUTE_IF("undo_space_read_fail", goto err_exit;);
-
     uint32_t id= mach_read_from_4(FIL_PAGE_SPACE_ID + page);
     if (id == 0 || id >= SRV_SPACE_ID_UPPER_BOUND ||
         memcmp_aligned<2>(FIL_PAGE_SPACE_ID + page,
                           FSP_HEADER_OFFSET + FSP_SPACE_ID + page, 4))
     {
-      sql_print_error("InnoDB: Inconsistent tablespace ID in file %s", name);
+      ib::error() << "Inconsistent tablespace ID in file " << name;
+      err= DB_CORRUPTION;
+      goto err_exit;
+    }
+
+    fsp_flags= mach_read_from_4(FSP_HEADER_OFFSET + FSP_SPACE_FLAGS + page);
+    if (buf_page_is_corrupted(false, page, fsp_flags))
+    {
+      ib::error() << "Checksum mismatch in the first page of file " << name;
+      err= DB_CORRUPTION;
       goto err_exit;
     }
 
     space_id= id;
-    fsp_flags= mach_read_from_4(FSP_HEADER_OFFSET + FSP_SPACE_FLAGS + page);
-
-    if (buf_page_is_corrupted(false, page, fsp_flags))
-    {
-      page= recv_sys.dblwr.find_page(page_id_t{space_id, 0}, LSN_MAX);
-      if (!page)
-      {
-        /* If the first page of the given user tablespace is not there
-        in the doublewrite buffer, then the recovery is going to fail
-        now. Report error only when doublewrite buffer is not empty */
-        sql_print_error("InnoDB: Corrupted page "
-                        "[page id: space=" UINT32PF ", page number=0]"
-                        " of datafile '%s' could not be found"
-                        " in the doublewrite buffer", space_id, name);
-        goto err_exit;
-      }
-
-      fsp_flags= mach_read_from_4(FSP_HEADER_OFFSET + FSP_SPACE_FLAGS + page);
-    }
+    aligned_free(page);
   }
 
   /* Load the tablespace into InnoDB's internal data structures. */
@@ -577,19 +558,20 @@ err_exit:
 
   fil_set_max_space_id_if_bigger(space_id);
 
-  mysql_mutex_lock(&fil_system.mutex);
-  fil_space_t *space= fil_space_t::create(uint32_t(space_id),
-                                          fsp_flags, false, nullptr,
-                                          FIL_ENCRYPTION_DEFAULT, true);
-  ut_ad(space);
+  fil_space_t *space= fil_space_t::create(space_id, fsp_flags,
+					  FIL_TYPE_TABLESPACE, NULL);
+  ut_a(fil_validate());
+  ut_a(space);
+
   fil_node_t *file= space->add(name, fh, 0, false, true);
+  mysql_mutex_lock(&fil_system.mutex);
 
   if (create)
   {
     space->set_sizes(SRV_UNDO_TABLESPACE_SIZE_IN_PAGES);
     space->size= file->size= uint32_t(size >> srv_page_size_shift);
   }
-  else if (!file->read_page0(page, true))
+  else if (!file->read_page0())
   {
     os_file_close(file->handle);
     file->handle= OS_FILE_CLOSED;
@@ -598,7 +580,6 @@ err_exit:
   }
 
   mysql_mutex_unlock(&fil_system.mutex);
-  aligned_free(apage);
   return space_id;
 }
 
@@ -624,6 +605,7 @@ srv_check_undo_redo_logs_exists()
 			OS_FILE_OPEN_RETRY
 			| OS_FILE_ON_ERROR_NO_EXIT
 			| OS_FILE_ON_ERROR_SILENT,
+			OS_FILE_NORMAL,
 			OS_DATA_FILE,
 			srv_read_only_mode,
 			&ret);
@@ -646,7 +628,7 @@ srv_check_undo_redo_logs_exists()
 	fh = os_file_create(innodb_log_file_key, logfilename.c_str(),
 			    OS_FILE_OPEN_RETRY | OS_FILE_ON_ERROR_NO_EXIT
 				    | OS_FILE_ON_ERROR_SILENT,
-			    OS_LOG_FILE, srv_read_only_mode,
+			    OS_FILE_NORMAL, OS_LOG_FILE, srv_read_only_mode,
 			    &ret);
 
 	if (ret) {
@@ -662,33 +644,32 @@ srv_check_undo_redo_logs_exists()
 	return(DB_SUCCESS);
 }
 
-static dberr_t srv_all_undo_tablespaces_open(bool create_new_db, ulint n_undo)
+static dberr_t srv_all_undo_tablespaces_open(bool create_new_db,
+                                             uint32_t n_undo)
 {
   /* Open all the undo tablespaces that are currently in use. If we
   fail to open any of these it is a fatal error. The tablespace ids
   should be contiguous. It is a fatal error because they are required
   for recovery and are referenced by the UNDO logs (a.k.a RBS). */
 
-  ulint prev_id= create_new_db ? srv_undo_space_id_start - 1 : 0;
+  uint32_t prev_id= create_new_db ? srv_undo_space_id_start - 1 : 0;
 
-  for (ulint i= 0; i < n_undo; ++i)
+  for (uint32_t i= 0; i < n_undo; ++i)
   {
     char name[OS_FILE_MAX_PATH];
-    snprintf(name, sizeof name, "%s/undo%03zu", srv_undo_dir, i + 1);
-    ulint space_id= srv_undo_tablespace_open(create_new_db, name, i);
-    switch (space_id) {
-    case ULINT_UNDEFINED:
-      return DB_CORRUPTION;
-    case 0:
+    snprintf(name, sizeof name, "%s/undo%03u", srv_undo_dir, i + 1);
+    uint32_t space_id= srv_undo_tablespace_open(create_new_db, name, i);
+    if (!space_id)
+    {
       if (!create_new_db)
-        goto unused_undo;
-      sql_print_error("InnoDB: Unable to open create tablespace '%s'", name);
+        break;
+      ib::error() << "Unable to open create tablespace '" << name << "'.";
       return DB_ERROR;
-    default:
-      /* Should be no gaps in undo tablespace ids. */
-      ut_a(!i || prev_id + 1 == space_id);
     }
 
+    /* Should be no gaps in undo tablespace ids. */
+    ut_a(!i || prev_id + 1 == space_id);
+
     prev_id= space_id;
 
     /* Note the first undo tablespace id in case of
@@ -701,14 +682,13 @@ static dberr_t srv_all_undo_tablespaces_open(bool create_new_db, ulint n_undo)
   We stop at the first failure. These are undo tablespaces that are
   not in use and therefore not required by recovery. We only check
   that there are no gaps. */
-unused_undo:
-  for (ulint i= prev_id + 1; i < srv_undo_space_id_start + TRX_SYS_N_RSEGS;
+
+  for (uint32_t i= prev_id + 1; i < srv_undo_space_id_start + TRX_SYS_N_RSEGS;
        ++i)
   {
      char name[OS_FILE_MAX_PATH];
-     snprintf(name, sizeof name, "%s/undo%03zu", srv_undo_dir, i);
-     ulint space_id= srv_undo_tablespace_open(create_new_db, name, i);
-     if (!space_id || space_id == ULINT_UNDEFINED)
+     snprintf(name, sizeof name, "%s/undo%03u", srv_undo_dir, i);
+     if (!srv_undo_tablespace_open(create_new_db, name, i))
        break;
      ++srv_undo_tablespaces_open;
   }
@@ -719,13 +699,12 @@ unused_undo:
 /** Open the configured number of dedicated undo tablespaces.
 @param[in]	create_new_db	whether the database is being initialized
 @return DB_SUCCESS or error code */
-dberr_t
-srv_undo_tablespaces_init(bool create_new_db)
+dberr_t srv_undo_tablespaces_init(bool create_new_db)
 {
   srv_undo_tablespaces_open= 0;
 
   ut_a(srv_undo_tablespaces <= TRX_SYS_N_RSEGS);
-  ut_a(!create_new_db || srv_operation <= SRV_OPERATION_EXPORT_RESTORED);
+  ut_a(!create_new_db || srv_operation == SRV_OPERATION_NORMAL);
 
   if (srv_undo_tablespaces == 1)
     srv_undo_tablespaces= 0;
@@ -756,34 +735,33 @@ srv_undo_tablespaces_init(bool create_new_db)
   already exist. */
   srv_undo_tablespaces_active= srv_undo_tablespaces;
 
-  ulint n_undo= (create_new_db || srv_operation == SRV_OPERATION_BACKUP ||
-                 srv_operation == SRV_OPERATION_RESTORE_DELTA)
+  uint32_t n_undo= (create_new_db || srv_operation == SRV_OPERATION_BACKUP ||
+                    srv_operation == SRV_OPERATION_RESTORE_DELTA)
     ? srv_undo_tablespaces : TRX_SYS_N_RSEGS;
 
-  mysql_mutex_lock(&recv_sys.mutex);
-  dberr_t err= srv_all_undo_tablespaces_open(create_new_db, n_undo);
-  mysql_mutex_unlock(&recv_sys.mutex);
+  if (dberr_t err= srv_all_undo_tablespaces_open(create_new_db, n_undo))
+    return err;
 
   /* Initialize srv_undo_space_id_start=0 when there are no
   dedicated undo tablespaces. */
   if (srv_undo_tablespaces_open == 0)
     srv_undo_space_id_start= 0;
 
-  if (err == DB_SUCCESS && create_new_db)
+  if (create_new_db)
   {
     mtr_t mtr;
-    for (ulint i= 0; i < srv_undo_tablespaces; ++i)
+    for (uint32_t i= 0; i < srv_undo_tablespaces; ++i)
     {
       mtr.start();
-      err= fsp_header_init(fil_space_get(srv_undo_space_id_start + i),
-                           SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, &mtr);
+      dberr_t err= fsp_header_init(fil_space_get(srv_undo_space_id_start + i),
+                                   SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, &mtr);
       mtr.commit();
       if (err)
-        break;
+        return err;
     }
   }
 
-  return err;
+  return DB_SUCCESS;
 }
 
 /** Create the temporary file tablespace.
@@ -843,19 +821,12 @@ srv_open_tmp_tablespace(bool create_new_db)
 	return(err);
 }
 
-/** Shutdown background threads, except the page cleaner.
-@param init_abort set to true when InnoDB startup aborted */
-static void srv_shutdown_threads(bool init_abort= false)
+/** Shutdown background threads, except the page cleaner. */
+static void srv_shutdown_threads()
 {
 	ut_ad(!srv_undo_sources);
 	srv_master_timer.reset();
-	/* In case of InnoDB start up aborted, Don't change
-	the srv_shutdown_state. Because innodb_shutdown()
-	does call innodb_preshutdown() which changes the
-	srv_shutdown_state back to SRV_SHUTDOWN_INITIATED */
-	if (!init_abort) {
-		srv_shutdown_state = SRV_SHUTDOWN_EXIT_THREADS;
-	}
+	srv_shutdown_state = SRV_SHUTDOWN_EXIT_THREADS;
 
 	if (purge_sys.enabled()) {
 		srv_purge_shutdown();
@@ -866,21 +837,6 @@ static void srv_shutdown_threads(bool init_abort= false)
 	}
 }
 
-
-/** Shut down background threads that can generate undo log. */
-static void srv_shutdown_bg_undo_sources()
-{
-  srv_shutdown_state= SRV_SHUTDOWN_INITIATED;
-
-  if (srv_undo_sources)
-  {
-    ut_ad(!srv_read_only_mode);
-    fts_optimize_shutdown();
-    dict_stats_shutdown();
-    srv_undo_sources= false;
-  }
-}
-
 #ifdef UNIV_DEBUG
 # define srv_init_abort(_db_err)	\
 	srv_init_abort_low(create_new_db, __FILE__, __LINE__, _db_err)
@@ -925,7 +881,7 @@ srv_init_abort_low(
 	}
 
 	srv_shutdown_bg_undo_sources();
-	srv_shutdown_threads(true);
+	srv_shutdown_threads();
 	return(err);
 }
 
@@ -937,7 +893,9 @@ static lsn_t srv_prepare_to_delete_redo_log_file(bool old_exists)
 {
   DBUG_ENTER("srv_prepare_to_delete_redo_log_file");
 
-  ut_ad(recv_sys.recovery_on);
+  /* Disable checkpoints in the page cleaner. */
+  ut_ad(!recv_sys.recovery_on);
+  recv_sys.recovery_on= true;
 
   /* Clean the buffer pool. */
   buf_flush_sync();
@@ -998,11 +956,7 @@ same_size:
   }
 
   ut_ad(flushed_lsn == log_sys.get_lsn());
-  ut_ad(!os_aio_pending_reads());
-  ut_d(mysql_mutex_lock(&buf_pool.flush_list_mutex));
-  ut_ad(!buf_pool.get_oldest_modification(0));
-  ut_d(mysql_mutex_unlock(&buf_pool.flush_list_mutex));
-  ut_d(os_aio_wait_until_no_pending_writes(false));
+  ut_ad(!buf_pool.any_io_pending());
 
   DBUG_RETURN(flushed_lsn);
 }
@@ -1091,21 +1045,17 @@ dberr_t srv_start(bool create_new_db)
 	bool		srv_log_file_found = true;
 	mtr_t		mtr;
 
-	ut_ad(srv_operation <= SRV_OPERATION_RESTORE_EXPORT
+	ut_ad(srv_operation == SRV_OPERATION_NORMAL
 	      || srv_operation == SRV_OPERATION_RESTORE
 	      || srv_operation == SRV_OPERATION_RESTORE_EXPORT);
 
 	if (srv_force_recovery) {
 		ib::info() << "!!! innodb_force_recovery is set to "
 			<< srv_force_recovery << " !!!";
-		if (srv_force_recovery == SRV_FORCE_NO_LOG_REDO) {
-			srv_read_only_mode = true;
-		}
 	}
 
-	if (srv_read_only_mode) {
-		sql_print_information("InnoDB: Started in read only mode");
-		srv_use_doublewrite_buf = false;
+	if (srv_force_recovery == SRV_FORCE_NO_LOG_REDO) {
+		srv_read_only_mode = true;
 	}
 
 	high_level_read_only = srv_read_only_mode
@@ -1155,6 +1105,12 @@ dberr_t srv_start(bool create_new_db)
 	mysql_stage_register("innodb", srv_stages,
 			     static_cast<int>(UT_ARR_SIZE(srv_stages)));
 
+	srv_max_n_threads =
+		1 /* dict_stats_thread */
+		+ 1 /* fts_optimize_thread */
+		+ 128 /* safety margin */
+		+ max_connections;
+
 	srv_boot();
 
 	ib::info() << my_crc32c_implementation();
@@ -1213,11 +1169,18 @@ dberr_t srv_start(bool create_new_db)
 		return(srv_init_abort(err));
 	}
 
-	if (srv_read_only_mode) {
+	srv_n_file_io_threads = srv_n_read_io_threads + srv_n_write_io_threads;
+
+	if (!srv_read_only_mode) {
+		/* Add the log and ibuf IO threads. */
+		srv_n_file_io_threads += 2;
+	} else {
 		ib::info() << "Disabling background log and ibuf IO write"
 			<< " threads.";
 	}
 
+	ut_a(srv_n_file_io_threads <= SRV_MAX_N_IO_THREADS);
+
 	if (os_aio_init()) {
 		ib::error() << "Cannot initialize AIO sub-system";
 
@@ -1262,12 +1225,7 @@ dberr_t srv_start(bool create_new_db)
 	}
 #endif /* UNIV_DEBUG */
 
-	if (!log_sys.create()) {
-		sql_print_error("InnoDB: Cannot allocate memory;"
-				" too large innodb_log_buffer_size?");
-		return srv_init_abort(DB_ERROR);
-	}
-
+	log_sys.create();
 	recv_sys.create();
 	lock_sys.create(srv_lock_table_size);
 
@@ -1280,60 +1238,61 @@ dberr_t srv_start(bool create_new_db)
 
 	/* Check if undo tablespaces and redo log files exist before creating
 	a new system tablespace */
-	srv_log_file_size_requested = srv_log_file_size;
-
-	if (innodb_encrypt_temporary_tables && !log_crypt_init()) {
-		return srv_init_abort(DB_ERROR);
-	}
-
-	/* Open system tablespace and undo spaces. It is a special case for
-	restore with empty redo log file. We support restore in such case and
-	proceed after opening system tablespaces. */
-	ulint	sum_of_new_sizes;
-
-	auto open_system_spaces = [&]() {
-		ut_ad(!create_new_db);
-		ut_ad(srv_operation == SRV_OPERATION_RESTORE ||
-		      srv_operation == SRV_OPERATION_RESTORE_EXPORT);
-
-		/* Open system tablespace ibdata1. */
-		auto err = srv_sys_space.open_or_create(
-			false, false, &sum_of_new_sizes, &flushed_lsn);
-		if (err != DB_SUCCESS) {
-			return srv_init_abort(err);
-		}
-
-		mysql_mutex_lock(&recv_sys.mutex);
-		bool all_opened = fil_system.sys_space->open(false);
-		mysql_mutex_unlock(&recv_sys.mutex);
-
-		/* Open data files for system tablespace. */
-		if (!all_opened) {
-			err = DB_ERROR;
-		} else {
-			/* Open undo tablespaces. */
-			err = srv_undo_tablespaces_init(false);
-		}
-		if (err != DB_SUCCESS) {
-			return srv_init_abort(err);
-		}
-		return DB_SUCCESS;
-	};
-
-	std::string logfile0;
-	bool create_new_log = create_new_db;
 	if (create_new_db) {
 		err = srv_check_undo_redo_logs_exists();
 		if (err != DB_SUCCESS) {
 			return(srv_init_abort(DB_ERROR));
 		}
 		recv_sys.debug_free();
+	}
+
+	/* Open or create the data files. */
+	ulint	sum_of_new_sizes;
+
+	err = srv_sys_space.open_or_create(
+		false, create_new_db, &sum_of_new_sizes, &flushed_lsn);
+
+	switch (err) {
+	case DB_SUCCESS:
+		break;
+	case DB_CANNOT_OPEN_FILE:
+		ib::error()
+			<< "Could not open or create the system tablespace. If"
+			" you tried to add new data files to the system"
+			" tablespace, and it failed here, you should now"
+			" edit innodb_data_file_path in my.cnf back to what"
+			" it was, and remove the new ibdata files InnoDB"
+			" created in this failed attempt. InnoDB only wrote"
+			" those files full of zeros, but did not yet use"
+			" them in any way. But be careful: do not remove"
+			" old data files which contain your precious data!";
+		/* fall through */
+	default:
+		/* Other errors might come from Datafile::validate_first_page() */
+		return(srv_init_abort(err));
+	}
+
+	srv_log_file_size_requested = srv_log_file_size;
+
+	if (innodb_encrypt_temporary_tables && !log_crypt_init()) {
+		return srv_init_abort(DB_ERROR);
+	}
+
+	std::string logfile0;
+	bool create_new_log = create_new_db;
+	if (create_new_db) {
 		flushed_lsn = log_sys.get_lsn();
 		log_sys.set_flushed_lsn(flushed_lsn);
 
 		err = create_log_file(true, flushed_lsn, logfile0);
 
 		if (err != DB_SUCCESS) {
+			for (Tablespace::const_iterator
+			       i = srv_sys_space.begin();
+			     i != srv_sys_space.end(); i++) {
+				os_file_delete(innodb_data_file_key,
+					       i->filepath());
+			}
 			return(srv_init_abort(err));
 		}
 	} else {
@@ -1342,96 +1301,63 @@ dberr_t srv_start(bool create_new_db)
 		bool log_file_found;
 		if (dberr_t err = find_and_check_log_file(log_file_found)) {
 			if (err == DB_NOT_FOUND) {
-				/* For restore, we need to continue. */
-				return open_system_spaces();
+				return DB_SUCCESS;
 			}
 			return srv_init_abort(err);
 		}
 
 		create_new_log = srv_log_file_size == 0;
-		if (!create_new_log) {
-			srv_log_file_found = log_file_found;
-
-			log_sys.log.open_file(get_log_file_path());
-
-			log_sys.log.create();
-
-			if (!log_set_capacity(
-				srv_log_file_size_requested)) {
-				return(srv_init_abort(DB_ERROR));
+		if (create_new_log) {
+			if (flushed_lsn < lsn_t(1000)) {
+				ib::error()
+					<< "Cannot create log file because"
+					" data files are corrupt or the"
+					" database was not shut down cleanly"
+					" after creating the data files.";
+				return srv_init_abort(DB_ERROR);
 			}
 
-			ut_ad(!recv_sys.recovery_on);
+			srv_log_file_size = srv_log_file_size_requested;
 
-			err= recv_recovery_read_max_checkpoint();
+			err = create_log_file(false, flushed_lsn, logfile0);
+
+			if (err == DB_SUCCESS) {
+				err = create_log_file_rename(flushed_lsn,
+							     logfile0);
+			}
 
 			if (err != DB_SUCCESS) {
-				return srv_init_abort(err);
+				return(srv_init_abort(err));
 			}
+
+			/* Suppress the message about
+			crash recovery. */
+			flushed_lsn = log_sys.get_lsn();
+			goto file_checked;
 		}
+
+		srv_log_file_found = log_file_found;
+
+		log_sys.log.open_file(get_log_file_path());
+
+		log_sys.log.create();
+
+		if (!log_set_capacity(srv_log_file_size_requested)) {
+			return(srv_init_abort(DB_ERROR));
+		}
+
+		/* Enable checkpoints in the page cleaner. */
+		recv_sys.recovery_on = false;
 	}
 
-	/* Open or create the data files. */
-	err = srv_sys_space.open_or_create(
-		false, create_new_db, &sum_of_new_sizes, &flushed_lsn);
-
-	switch (err) {
-	case DB_SUCCESS:
-		break;
-	case DB_CANNOT_OPEN_FILE:
-		sql_print_error("InnoDB: Could not open or create the system"
-			" tablespace. If you tried to add new data files"
-			" to the system tablespace, and it failed here,"
-			" you should now edit innodb_data_file_path"
-			" in my.cnf back to what it was, and remove the"
-			" new ibdata files InnoDB created in this failed"
-			" attempt. InnoDB only wrote those files full of"
-			" zeros, but did not yet use them in any way. But"
-			" be careful: do not remove old data files which"
-			" contain your precious data!");
-		/* fall through */
-	default:
-		/* Other errors might come from Datafile::validate_first_page() */
-		return(srv_init_abort(err));
-	}
-
-	if (!create_new_db && create_new_log) {
-		if (flushed_lsn < lsn_t(1000)) {
-			sql_print_error(
-				"InnoDB: Cannot create log file because"
-				" data files are corrupt or the"
-				" database was not shut down cleanly"
-				" after creating the data files.");
-			return srv_init_abort(DB_ERROR);
-		}
-
-		srv_log_file_size = srv_log_file_size_requested;
-
-		err = create_log_file(false, flushed_lsn, logfile0);
-		if (err == DB_SUCCESS) {
-			err = create_log_file_rename(flushed_lsn, logfile0);
-		}
-
-		if (err != DB_SUCCESS) {
-			return(srv_init_abort(err));
-		}
-
-		/* Suppress the message about
-		crash recovery. */
-		flushed_lsn = log_sys.get_lsn();
-	}
-
-	/* Open data files in the system tablespace: we keep
-	them open until database shutdown */
-	mysql_mutex_lock(&recv_sys.mutex);
+file_checked:
+	/* Open log file and data files in the systemtablespace: we keep
+        them open until database shutdown */
 	ut_d(fil_system.sys_space->recv_size = srv_sys_space_size_debug);
-	err = fil_system.sys_space->open(create_new_db)
-		? DB_SUCCESS : DB_ERROR;
-	mysql_mutex_unlock(&recv_sys.mutex);
 
-	if (err == DB_SUCCESS) {
-		err = srv_undo_tablespaces_init(create_new_db);
-	}
+	err = fil_system.sys_space->open(create_new_db)
+		? srv_undo_tablespaces_init(create_new_db)
+		: DB_ERROR;
 
 	/* If the force recovery is set very high then we carry on regardless
 	of all errors. Basically this is fingers crossed mode. */
@@ -1516,13 +1442,14 @@ dberr_t srv_start(bool create_new_db)
 			: recv_recovery_from_checkpoint_start(flushed_lsn);
 		recv_sys.close_files();
 
+		recv_sys.dblwr.pages.clear();
+
 		if (err != DB_SUCCESS) {
 			return(srv_init_abort(err));
 		}
 
 		switch (srv_operation) {
-                case SRV_OPERATION_NORMAL:
-		case SRV_OPERATION_EXPORT_RESTORED:
+		case SRV_OPERATION_NORMAL:
 		case SRV_OPERATION_RESTORE_EXPORT:
 			/* Initialize the change buffer. */
 			err = dict_boot();
@@ -1555,14 +1482,9 @@ dberr_t srv_start(bool create_new_db)
 		if (srv_force_recovery < SRV_FORCE_NO_LOG_REDO) {
 			/* Apply the hashed log records to the
 			respective file pages, for the last batch of
-			recv_group_scan_log_recs().
-			Since it may generate huge batch of threadpool tasks,
-			for read io task group, scale down thread creation rate
-			by temporarily restricting tpool concurrency.
-			*/
-			srv_thread_pool->set_concurrency(srv_n_read_io_threads);
+			recv_group_scan_log_recs(). */
+
 			recv_sys.apply(true);
-			srv_thread_pool->set_concurrency();
 
 			if (recv_sys.is_corrupt_log()
 			    || recv_sys.is_corrupt_fs()) {
@@ -1578,118 +1500,10 @@ dberr_t srv_start(bool create_new_db)
 
 		fil_system.space_id_reuse_warned = false;
 
-		if (srv_operation == SRV_OPERATION_RESTORE
-		    || srv_operation == SRV_OPERATION_RESTORE_EXPORT) {
-			buf_flush_sync();
-			recv_sys.debug_free();
-			/* After applying the redo log from
-			SRV_OPERATION_BACKUP, flush the changes
-			to the data files and truncate or delete the log.
-			Unless --export is specified, no further change to
-			InnoDB files is needed. */
-			ut_ad(srv_force_recovery <= SRV_FORCE_IGNORE_CORRUPT);
-			ut_ad(recv_no_log_write);
-			err = fil_write_flushed_lsn(log_sys.get_lsn());
-			ut_ad(!os_aio_pending_reads());
-			ut_d(mysql_mutex_lock(&buf_pool.flush_list_mutex));
-			ut_ad(!buf_pool.get_oldest_modification(0));
-			ut_d(mysql_mutex_unlock(&buf_pool.flush_list_mutex));
-			log_sys.log.close_file();
-			if (err == DB_SUCCESS) {
-				bool trunc = srv_operation
-					== SRV_OPERATION_RESTORE;
-				if (!trunc) {
-					delete_log_file("0");
-				} else {
-					auto logfile0 = get_log_file_path();
-					/* Truncate the first log file. */
-					fclose(fopen(logfile0.c_str(), "w"));
-				}
-			}
-			return(err);
-		}
-		/* Upgrade or resize or rebuild the redo logs before
-		generating any dirty pages, so that the old redo log
-		file will not be written to. */
-
-		if (srv_force_recovery == SRV_FORCE_NO_LOG_REDO) {
-			/* Completely ignore the redo log. */
-		} else if (srv_read_only_mode) {
-			/* Leave the redo log alone. */
-		} else if (srv_log_file_size_requested == srv_log_file_size
-			   && srv_log_file_found
-			   && log_sys.log.format
-			   == (srv_encrypt_log
-			       ? log_t::FORMAT_ENC_10_5
-			       : log_t::FORMAT_10_5)
-			   && log_sys.log.subformat == 2) {
-			/* No need to add or remove encryption,
-			upgrade, downgrade, or resize. */
-		} else {
-			/* Prepare to delete the old redo log file */
-			flushed_lsn = srv_prepare_to_delete_redo_log_file(
-				srv_log_file_found);
-
-			DBUG_EXECUTE_IF("innodb_log_abort_1",
-					return(srv_init_abort(DB_ERROR)););
-			/* Prohibit redo log writes from any other
-			threads until creating a log checkpoint at the
-			end of create_log_file(). */
-			ut_d(recv_no_log_write = true);
-			ut_ad(!os_aio_pending_reads());
-			ut_d(mysql_mutex_lock(&buf_pool.flush_list_mutex));
-			ut_ad(!buf_pool.get_oldest_modification(0));
-			ut_d(mysql_mutex_unlock(&buf_pool.flush_list_mutex));
-			/* os_aio_pending_writes() may hold here if
-			some write_io_callback() did not release the
-			slot yet. However, the page write itself must
-			have completed, because the buf_pool.flush_list
-			is empty. In debug builds, we wait for this to
-			happen, hoping to get a hung process if this
-			assumption does not hold. */
-			ut_d(os_aio_wait_until_no_pending_writes(false));
-
-			DBUG_EXECUTE_IF("innodb_log_abort_3",
-					return(srv_init_abort(DB_ERROR)););
-			DBUG_PRINT("ib_log", ("After innodb_log_abort_3"));
-
-			/* Stamp the LSN to the data files. */
-			err = fil_write_flushed_lsn(flushed_lsn);
-
-			DBUG_EXECUTE_IF("innodb_log_abort_4", err = DB_ERROR;);
-			DBUG_PRINT("ib_log", ("After innodb_log_abort_4"));
-
-			if (err != DB_SUCCESS) {
-				return(srv_init_abort(err));
-			}
-
-			/* Close the redo log file, so that we can replace it */
-			log_sys.log.close_file();
-
-			DBUG_EXECUTE_IF("innodb_log_abort_5",
-					return(srv_init_abort(DB_ERROR)););
-			DBUG_PRINT("ib_log", ("After innodb_log_abort_5"));
-
-			ib::info()
-				<< "Starting to delete and rewrite log file.";
-
-			srv_log_file_size = srv_log_file_size_requested;
-			err = create_log_file(false, flushed_lsn, logfile0);
-
-			if (err == DB_SUCCESS) {
-				err = create_log_file_rename(flushed_lsn,
-							     logfile0);
-			}
-
-			if (err != DB_SUCCESS) {
-				return(srv_init_abort(err));
-			}
-		}
-		recv_sys.debug_free();
-
 		if (!srv_read_only_mode) {
-			const ulint flags = FSP_FLAGS_PAGE_SSIZE();
-			for (ulint id = 0; id <= srv_undo_tablespaces; id++) {
+			const uint32_t flags = FSP_FLAGS_PAGE_SSIZE();
+			for (uint32_t id = 0; id <= srv_undo_tablespaces;
+			     id++) {
 				if (fil_space_t* space = fil_space_get(id)) {
 					fsp_flags_try_adjust(space, flags);
 				}
@@ -1770,6 +1584,102 @@ dberr_t srv_start(bool create_new_db)
 				return(srv_init_abort(DB_ERROR));
 			}
 		}
+
+		recv_sys.debug_free();
+
+		if (srv_operation == SRV_OPERATION_RESTORE
+		    || srv_operation == SRV_OPERATION_RESTORE_EXPORT) {
+			/* After applying the redo log from
+			SRV_OPERATION_BACKUP, flush the changes
+			to the data files and truncate or delete the log.
+			Unless --export is specified, no further change to
+			InnoDB files is needed. */
+			ut_ad(srv_force_recovery <= SRV_FORCE_IGNORE_CORRUPT);
+			ut_ad(recv_no_log_write);
+			err = fil_write_flushed_lsn(log_sys.get_lsn());
+			DBUG_ASSERT(!buf_pool.any_io_pending());
+			log_sys.log.close_file();
+			if (err == DB_SUCCESS) {
+				bool trunc = srv_operation
+					== SRV_OPERATION_RESTORE;
+				if (!trunc) {
+					delete_log_file("0");
+				} else {
+					auto logfile0 = get_log_file_path();
+					/* Truncate the first log file. */
+					fclose(fopen(logfile0.c_str(), "w"));
+				}
+			}
+			return(err);
+		}
+
+		/* Upgrade or resize or rebuild the redo logs before
+		generating any dirty pages, so that the old redo log
+		file will not be written to. */
+
+		if (srv_force_recovery == SRV_FORCE_NO_LOG_REDO) {
+			/* Completely ignore the redo log. */
+		} else if (srv_read_only_mode) {
+			/* Leave the redo log alone. */
+		} else if (srv_log_file_size_requested == srv_log_file_size
+			   && srv_log_file_found
+			   && log_sys.log.format
+			   == (srv_encrypt_log
+			       ? log_t::FORMAT_ENC_10_5
+			       : log_t::FORMAT_10_5)
+			   && log_sys.log.subformat == 2) {
+			/* No need to add or remove encryption,
+			upgrade, downgrade, or resize. */
+		} else {
+			/* Prepare to delete the old redo log file */
+			flushed_lsn = srv_prepare_to_delete_redo_log_file(
+				srv_log_file_found);
+
+			DBUG_EXECUTE_IF("innodb_log_abort_1",
+					return(srv_init_abort(DB_ERROR)););
+			/* Prohibit redo log writes from any other
+			threads until creating a log checkpoint at the
+			end of create_log_file(). */
+			ut_d(recv_no_log_write = true);
+			DBUG_ASSERT(!buf_pool.any_io_pending());
+
+			DBUG_EXECUTE_IF("innodb_log_abort_3",
+					return(srv_init_abort(DB_ERROR)););
+			DBUG_PRINT("ib_log", ("After innodb_log_abort_3"));
+
+			/* Stamp the LSN to the data files. */
+			err = fil_write_flushed_lsn(flushed_lsn);
+
+			DBUG_EXECUTE_IF("innodb_log_abort_4", err = DB_ERROR;);
+			DBUG_PRINT("ib_log", ("After innodb_log_abort_4"));
+
+			if (err != DB_SUCCESS) {
+				return(srv_init_abort(err));
+			}
+
+			/* Close the redo log file, so that we can replace it */
+			log_sys.log.close_file();
+
+			DBUG_EXECUTE_IF("innodb_log_abort_5",
+					return(srv_init_abort(DB_ERROR)););
+			DBUG_PRINT("ib_log", ("After innodb_log_abort_5"));
+
+			ib::info()
+				<< "Starting to delete and rewrite log file.";
+
+			srv_log_file_size = srv_log_file_size_requested;
+
+			err = create_log_file(false, flushed_lsn, logfile0);
+
+			if (err == DB_SUCCESS) {
+				err = create_log_file_rename(flushed_lsn,
+							     logfile0);
+			}
+
+			if (err != DB_SUCCESS) {
+				return(srv_init_abort(err));
+			}
+		}
 	}
 
 	ut_ad(err == DB_SUCCESS);
@@ -1869,13 +1779,21 @@ dberr_t srv_start(bool create_new_db)
 		}
 
 		if (srv_force_recovery < SRV_FORCE_NO_UNDO_LOG_SCAN) {
-			/* The following call is necessary for the change
+			/* The following call is necessary for the insert
 			buffer to work with multiple tablespaces. We must
 			know the mapping between space id's and .ibd file
 			names.
 
+			In a crash recovery, we check that the info in data
+			dictionary is consistent with what we already know
+			about space id's from the calls to fil_ibd_load().
+
+			In a normal startup, we create the space objects for
+			every table in the InnoDB data dictionary that has
+			an .ibd file.
+
 			We also determine the maximum tablespace id used. */
-			dict_check_tablespaces_and_store_max_id(nullptr);
+			dict_check_tablespaces_and_store_max_id();
 		}
 
 		if (srv_force_recovery < SRV_FORCE_NO_TRX_UNDO
@@ -1926,8 +1844,7 @@ skip_monitors:
 		return(srv_init_abort(err));
 	}
 
-	if (!srv_read_only_mode
-	    && srv_operation <= SRV_OPERATION_EXPORT_RESTORED) {
+	if (!srv_read_only_mode && srv_operation == SRV_OPERATION_NORMAL) {
 		/* Initialize the innodb_temporary tablespace and keep
 		it open until shutdown. */
 		err = srv_open_tmp_tablespace(create_new_db);
@@ -1996,6 +1913,19 @@ skip_monitors:
 	return(DB_SUCCESS);
 }
 
+/** Shut down background threads that can generate undo log. */
+void srv_shutdown_bg_undo_sources()
+{
+	srv_shutdown_state = SRV_SHUTDOWN_INITIATED;
+
+	if (srv_undo_sources) {
+		ut_ad(!srv_read_only_mode);
+		fts_optimize_shutdown();
+		dict_stats_shutdown();
+		srv_undo_sources = false;
+	}
+}
+
 /**
   Shutdown purge to make sure that there is no possibility that we call any
   plugin code (e.g., audit) inside virtual column computation.
@@ -2009,13 +1939,13 @@ void innodb_preshutdown()
 
   if (srv_read_only_mode)
     return;
-  if (!srv_fast_shutdown && srv_operation <= SRV_OPERATION_EXPORT_RESTORED)
+  if (!srv_fast_shutdown && srv_operation == SRV_OPERATION_NORMAL)
   {
     /* Because a slow shutdown must empty the change buffer, we had
     better prevent any further changes from being buffered. */
     innodb_change_buffering= 0;
 
-    if (srv_force_recovery < SRV_FORCE_NO_TRX_UNDO && srv_was_started)
+    if (trx_sys.is_initialised())
       while (trx_sys.any_active_transactions())
         std::this_thread::sleep_for(std::chrono::milliseconds(1));
   }
@@ -2049,7 +1979,6 @@ void innodb_shutdown()
 		mysql_mutex_unlock(&buf_pool.flush_list_mutex);
 		break;
 	case SRV_OPERATION_NORMAL:
-	case SRV_OPERATION_EXPORT_RESTORED:
 		/* Shut down the persistent files. */
 		logs_empty_and_mark_files_at_shutdown();
 	}
@@ -2076,13 +2005,12 @@ void innodb_shutdown()
 
 	ut_ad(dict_sys.is_initialised() || !srv_was_started);
 	ut_ad(trx_sys.is_initialised() || !srv_was_started);
-	ut_ad(buf_dblwr.is_created() || !srv_was_started
+	ut_ad(buf_dblwr.is_initialised() || !srv_was_started
 	      || srv_read_only_mode
 	      || srv_force_recovery >= SRV_FORCE_NO_TRX_UNDO);
 	ut_ad(lock_sys.is_initialised() || !srv_was_started);
 	ut_ad(log_sys.is_initialised() || !srv_was_started);
-	ut_ad(ibuf.index || !innodb_change_buffering || !srv_was_started
-	      || srv_force_recovery >= SRV_FORCE_NO_DDL_UNDO);
+	ut_ad(ibuf.index || !srv_was_started);
 
 	dict_stats_deinit();
 
@@ -2135,6 +2063,10 @@ void innodb_shutdown()
 	}
 	srv_tmp_space.shutdown();
 
+	if (srv_stats.pages_page_compression_error)
+		ib::warn() << "Page compression errors: "
+			   << srv_stats.pages_page_compression_error;
+
 	if (srv_was_started && srv_print_verbose_log) {
 		ib::info() << "Shutdown completed; log sequence number "
 			   << srv_shutdown_lsn
